#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

void Show(char *s);

void StepReset();

void StepShow();

char *ShowDots(int dotwatch);

int  StateReport = 1;
void StateShow();

void CastRings();

int  IsVowel(char c);

/*------------------------------------*/

#define UP_DOT  1
#define NO_DOT  2
#define DN_DOT  3

int r1[26];
int r2[26];
int r3[26];

int d[3]={UP_DOT,NO_DOT,DN_DOT};

char showd[4]  = {'!','\'','-','.'};
char showa[26] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

/* Ring State variables */

int s1=0;
int s2=0;
int s3=0;

int step = 0;
int monitor;

char dots[4];

void CastRings() {

	int i;

	for (i=0; i<26; i++ ) {
		r1[i] = 65 + i + (d[(i % 3)/1]<<8);
		r2[i] = 65 + i + (d[(i % 6)/2]<<8);
		r3[i] = 65 + i + (d[(i % 9)/3]<<8);
		}
	}

int IsVowel(char c) {
   return ((c=='A')||(c=='E')||(c=='I')||(c=='O')||(c=='U'));
   }



void StepReset() {
	step = 0;
	}

void StepShow() {
	if (StateReport) printf("\n%03d ",++step);
	}

void StateShow() {

	char *monresult;
	char monstring[2] = "-";;

	if (StateReport) {
		if (monitor > 255) {
			 monresult = ShowDots(monitor);
			 }
		if (monitor < 256) {
			 monresult = &(monstring[0]);
			 monstring[0] = monitor;
			 }

		printf("\n[%c%c %c%c %c%c|<%s]",
			showd[r1[s1]>>8], r1[s1] & 0xFF,
			showd[r2[s2]>>8], r2[s2] & 0xFF,
			showd[r3[s3]>>8], r3[s3] & 0xFF,

			monresult );
		}
	}

void Show(char *s) {
	if (StateReport) printf("\n[%s>\n",s);
	}

char *ShowDots(int dotwatch) {
   int dotc = dotwatch >> 8;

   int dot1 = (dotwatch >> 4) & 3;
   int dot2 = (dotwatch >> 2) & 3;
   int dot3 = (dotwatch     ) & 3;

   if (dotc == 0) {
	  dots[0] = 0;
	  }
   if (dotc == 1) {
	  dots[0] = showd[dot3];
	  dots[1] = 0;
	  }
   if (dotc == 2) {
	  dots[0] = showd[dot2];
	  dots[1] = showd[dot3];
	  dots[2] = 0;
	  }
   if (dotc == 3) {
	  dots[0] = showd[dot1];
	  dots[1] = showd[dot2];
	  dots[2] = showd[dot3];
	  dots[3] = 0;
	  }

   return &(dots[0]);
   }


int nstate() {
   return s1*676+s2*26+s3;
   }


int getring(int a) {
   if (a == 1) return r1[s1] & 0xFF;
   if (a == 2) return r2[s2] & 0xFF;
   if (a == 3) return r3[s3] & 0xFF;
   return '@';
   }

int getdots(int a, int b, int c) {
   int dot1 = 0;
   int dot2 = 0;
   int dot3 = 0;
   int dotc = 0;
   int dotsum;

   if (c == 0) { c=b; b=a; a=0; }
   if (c == 0) { c=b; b=a; a=0; }
   if (c == 0) { c=b; b=a; a=0; }

   if (a == 1) dot1 = r1[s1]>>8;
   if (a == 2) dot1 = r2[s2]>>8;
   if (a == 3) dot1 = r3[s3]>>8;
   if (b == 1) dot2 = r1[s1]>>8;
   if (b == 2) dot2 = r2[s2]>>8;
   if (b == 3) dot2 = r3[s3]>>8;
   if (c == 1) dot3 = r1[s1]>>8;
   if (c == 2) dot3 = r2[s2]>>8;
   if (c == 3) dot3 = r3[s3]>>8;

   if (a != 0) { dotc++; }
   if (b != 0) { dotc++; }
   if (c != 0) { dotc++; }

   dotsum = (dot3)+(dot2<<2)+(dot1<<4)+(dotc<<8);

   return dotsum;
   }


#define R(x)   ((x)&0xFF)
#define D(x)   (dots[(x)>>8])

void Ring1Up()  {s1=(s1+ 1)%26;}
void Ring1Dn()  {s1=(s1+25)%26;}
void Ring2Up()  {s2=(s2+ 1)%26;}
void Ring2Dn()  {s2=(s2+25)%26;}
void Ring3Up()  {s3=(s3+ 1)%26;}
void Ring3Dn()  {s3=(s3+25)%26;}
void Ring23Up() {s2=(s2+ 1)%26; s3=(s3+ 1)%26;}
void Ring23Dn() {s2=(s2+25)%26; s3=(s3+25)%26;}
void RingUp()   {s1=(s1+ 1)%26; s2=(s2+ 1)%26; s3=(s3+ 1)%26;}
void RingDn()   {s1=(s1+25)%26; s2=(s2+25)%26; s3=(s3+25)%26;}

char Ring(int n) {
	if (n==1) return r1[s1]&0xFF;
	if (n==2) return r2[s2]&0xFF;
	if (n==3) return r3[s3]&0xFF;
	return 0;
	}


#define Ring1      (s1)
#define Ring2      (s2)
#define Ring3      (s3)
#define Rings1and2 ((Ring1)*256+(Ring2))
#define Rings2and3 ((Ring2)*256+(Ring3))

#define Dots(a,b,c) getdots(a,b,c)

#define LookAt(something)    {monitor=(something);}

#define TurnRing2UpUntilMatch(something)       {P("T2u");do{S;R2u;}while(monitor!=(something));S;}
#define TurnRing3UpUntilMatch(something)       {P("T3u");do{S;R3u;}while(monitor!=(something));S;}
#define TurnRings2and3UpUntilMatch(something)  {P("T23u");do{S;R23u;}while(monitor!=(something));S;}

#define TurnWholeRingUpUntilMatch(something)   {P("TRu");do{S;Ru;}while(monitor!=(something));S;}

#define TurnRing2DownUntilMatch(something)       (P("T2d");do{R2d;}while(monitor!=(something)))
#define TurnRing3DownUntilMatch(something)       (P("T3d");do{R3d;}while(monitor!=(something)))
#define TurnRings2and3DownUntilMatch(something)  (P("T23d");do{R2d;R3d;}while(monitor!=(something)))

#define TurnWholeRingDownUntilMatch(something)  (P("TRd");do{Rd;}while(monitor!=(something)))

void setrings(int n1, int n2, int n3) {
   s1 = ((n1 & 0x1F)-1) % 26;
   s2 = ((n2 & 0x1F)-1) % 26;
   s3 = ((n3 & 0x1F)-1) % 26;
   }

void SetRingsA(char a, char b, char c) {
   if (a) s1 = (a-'A') % 26;
   if (b) s2 = (b-'A') % 26;
   if (c) s3 = (c-'A') % 26;
   }

void setr1(int n) { s1 = n % 26; }
void setr2(int n) { s2 = n % 26; }
void setr3(int n) { s3 = n % 26; }

void setr1a(char a) { s1 = (a-'A') % 26; }
void setr2a(char a) { s2 = (a-'A') % 26; }
void setr3a(char a) { s3 = (a-'A') % 26; }



char EncodeKeyAndChar(char nextkey, char nextchar) {

   char *msg = "[ENC:-->";

   msg[5] = nextkey;
   msg[6] = nextchar;

   Show(msg);

   monitor = getdots(0,1,2);

   StateShow();

   if (nextkey <= 'M') {
	   Show("<dn>");
	   do {
		   RingDn();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   } else {
	   Show("<up>");
	   do {
		   RingUp();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   }

   Show("<found>");

   monitor = 0;

   setr3a(nextkey);

   StateShow();

   /* reverse direction from last time */

   monitor = getdots(0,1,2);

   StateShow();

   if (nextkey > 'M') {
	   Show("<dn>");
	   do {
		   RingDn();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   } else {
	   Show("<up>");
	   do {
		   RingUp();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   }


   Show("<found>");

   /* Must be this way to allow both encode, decode */
   /* Letter match means direction does not matter  */

   monitor = nextchar;
   Show("<dn>");
   StateShow();

   do {
	   Ring23Dn();
	   StateShow();
   }
   while (monitor != Ring(2));

   /* Must be last step to provide encoded char that */
   /* can be decoded                                 */

   monitor = 0;

   StateShow();

   Show("<ENC]");

   return Ring(3);
   }


char DecodeKeyAndChar(char nextkey, char nextchar) {
   char *msg = "[DEC:-->";

   msg[5] = nextkey;
   msg[6] = nextchar;

   Show(msg);

   monitor = getdots(0,1,2);

   StateShow();

   if (nextkey <= 'M') {
	   Show("<dn>");
	   do {
		   RingDn();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   } else {
	   Show("<up>");
	   do {
		   RingUp();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   }

   Show("<found>");

   monitor = 0;

   setr3a(nextkey);

   StateShow();

   /* reverse direction from last time */

   monitor = getdots(0,1,2);

   StateShow();

   if (nextkey > 'M') {
	   Show("<dn>");
	   do {
		   RingDn();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   } else {
	   Show("<up>");
	   do {
		   RingUp();
		   StateShow();
	   }
	   while (getdots(0,2,3) != monitor);
   }

   Show("<found>");

   /* Must be this way to allow both encode, decode */
   /* Letter match means direction does not matter  */

   monitor = nextchar;
   Show("<dn>");
   StateShow();

   do {
	   Ring23Dn();
	   StateShow();
   }
   while (monitor != Ring(3));

   /* Must be last step to provide encoded char that */
   /* can be decoded                                 */

   monitor = 0;

   StateShow();

   Show("<DEC]");

   return Ring(2);
   }



int main(int argc, char *argv[]) {

   int doDecode = 0;

   char *encode;
   char *keystore;
   char *key;
   char *result;

   if (argc <= 2) {
	   printf("Fidget, the secret decoder ring protocol\n");
	   printf("\n");
	   printf("encode:   fidget  key stringtoencode\n");
	   printf("decode:   fidget -key stringtodecode\n");
	   printf("\n");
	   printf("All strings must be uppercase letters.\n");
	   printf("\n");
	   printf("Prefix encode or decode string with '?' to\n");
	   printf("all details of ring movements.\n");
	   return;
	   }

	if (argc > 2) keystore = argv[1];
	if (argc > 2) encode = argv[2];

	if (keystore[0] == '-') {
		doDecode = 1;
		keystore = &keystore[1];
		}

	key = keystore;

	StateReport = 0;
	if (*encode == '?') {
		StateReport = 1;
		encode++;
		}

	result = encode;

	CastRings();

	setrings(key[0], key[1], key[2]);

	key++; key++; key++;

	if (*key == 0) key = keystore;

	while (isupper(*encode)) {

		printf("%c%c%c %c %c ",Ring(1),Ring(2),Ring(3),*key,*encode);

		if (doDecode) {
			*encode = DecodeKeyAndChar( *key++, *encode );
		} else {
			*encode = EncodeKeyAndChar( *key++, *encode );
		}
		encode++;
		if (*key == 0) key = keystore;

		printf("%c%c%c\n",Ring(1),Ring(2),Ring(3));
	}
   printf("Final: %s\n", result);
   }
